"""Setup script for the `wsgi-statsd' package."""

import codecs
import sys
from os.path import abspath, dirname, join

from setuptools import setup
from setuptools.command.test import test as TestCommand

import wsgi_statsd


class ToxTestCommand(TestCommand):

    """Test command which runs tox under the hood."""

    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        """Initialize options and set their defaults."""
        TestCommand.initialize_options(self)
        self.tox_args = '--recreate'

    def finalize_options(self):
        """Add options to the test runner (tox)."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Invoke the test runner (tox)."""
        # import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        errno = tox.cmdline(args=shlex.split(self.tox_args))
        sys.exit(errno)

long_description = []

for text_file in ['README.rst', 'CHANGES.rst']:
    with codecs.open(join(dirname(abspath(__file__)), text_file), encoding='utf-8') as f:
        long_description.append(f.read())


setup(
    name="wsgi-statsd",
    description="WSGI middleware for statsd timing",
    long_description='\n'.join(long_description),
    author="Wouter Lansu, Paylogic International and others",
    license="MIT license",
    author_email="developers@paylogic.com",
    url="https://github.com/paylogic/wsgi-statsd",
    version=wsgi_statsd.__version__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware'
    ],
    cmdclass={"test": ToxTestCommand},
    packages=["wsgi_statsd"],
    install_requires=['statsd'],
    tests_require=["tox"],
)



"""StatsdTimingMiddleware object."""
import time

__version__ = '0.3.1'

import re

CHAR_RE = re.compile(r'[^\w]')


class StatsdTimingMiddleware(object):

    """The Statsd timing middleware."""

    def __init__(self, app, client, time_exceptions=False, separator='_'):
        """Initialize the middleware with an application and a Statsd client.

        :param app: The application object.
        :param client: `statsd.StatsClient` object.
        :param time_exceptions: send stats when exception happens or not, `False` by default.
        :type time_exceptions: bool
        :param separator: separator of the parts of key sent to statsd, defaults to '_'
        :type separator: str
        """
        self.app = app
        self.statsd_client = client
        self.time_exceptions = time_exceptions
        self.separator = separator

    def __call__(self, environ, start_response):
        """Call the application and time it.

        :param environ: Dictionary object, containing CGI-style environment variables.
        :param start_response: Callable used to begin the HTTP response.
        """
        response_interception = {}

        def start_response_wrapper(status, response_headers, exc_info=None):
            """Wrap the start_response in order to retrieve the status code."""
            response_interception.update(status=status, response_headers=response_headers, exc_info=exc_info)
            return start_response(status, response_headers, exc_info)

        # Time the call.
        start = time.time()
        try:
            response = self.app(environ, start_response_wrapper)
            try:
                for event in response:
                    yield event
            finally:
                if hasattr(response, 'close'):
                    response.close()
            self.send_stats(start, environ, response_interception)
        except Exception as exception:
            if self.time_exceptions:
                self.send_stats(start, environ, response_interception, exception)
            raise

    def get_key_name(self, environ, response_interception, exception=None):
        """Get the timer key name.

        :param environ: wsgi environment
        :type environ: dict
        :param response_interception: dictionary in form
            {'status': '<response status>', 'response_headers': [<response headers], 'exc_info': <exc_info>}
            This is the interception of what was passed to start_response handler.
        :type response_interception: dict
        :param exception: optional exception happened during the iteration of the response
        :type exception: Exception

        :return: string in form '{{UNDERSCORED_PATH}}.{{METHOD}}.{{STATUS_CODE}}'
        :rtype: str
        """
        status = response_interception.get('status')
        status_code = status.split()[0]  # Leave only the status code.
        # PATH_INFO can be empty, so falling back to '/' in that case
        path = CHAR_RE.sub(self.separator, (environ['PATH_INFO'].rstrip('\/') or '/')[1:])
        parts = [path, environ['REQUEST_METHOD'], status_code]
        if exception:
            parts.append(exception.__class__.__name__)
        return '.'.join(parts)

    def send_stats(self, start, environ, response_interception, exception=None):
        """Send the actual timing stats.

        :param start: start time in seconds since the epoch as a floating point number
        :type start: float
        :param environ: wsgi environment
        :type environ: dict
        :param response_interception: dictionary in form
            {'status': '<response status>', 'response_headers': [<response headers], 'exc_info': <exc_info>}
            This is the interception of what was passed to start_response handler.
        :type response_interception: dict
        :param exception: optional exception happened during the iteration of the response
        :type exception: Exception
        """
        # It could happen that start_response wasn't called or it failed, so we might have an empty interception
        if response_interception:
            # Create the timer object and send the data to statsd.
            key_name = self.get_key_name(environ, response_interception, exception=exception)
            timer = self.statsd_client.timer(key_name)
            timer._start_time = start
            timer.stop()
